/* httpsync.c version 1.12  Copyright 1999 Forrest J. Cavalier III

   See http://www.mibsoftware.com/httpsync/ for documentation and
   original copies of this software.

   httpsync version 1.12 is provided by Forrest J. Cavalier III,
   d-b-a Mib Software, under the following terms and conditions,
   which are believed to meet the open-source definition,
   version 1.0 at http://www.opensource.org

   Contact Mib Software (mibsoft@mibsoftware.com) to discuss custom
   and enhanced versions of this and other software.  We would appreciate
   that you notify us of defect discoveries and enhancements so that
   others can benefit from work on open-source software.

  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software (the "Software"), to deal in the Software
   without restriction, including without limitation the rights to use,
   copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software
   is furnished to do so, subject to the following conditions:

   1. The above copyright notice and this permission notice shall
      be included in the source code of all copies or substantial
      portions of the Software.

   2. Modifications to the source code may be created and the software
      built from modified source code may be distributed.  Modifications
      distributed as source code shall be distributed only as "patch files."
      When source code of the Software is distributed, it shall be in its
      "pristine", unmodified form.

   3. Modified works that are distributed must be named differently
      than the original.

   4. THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
      EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
      MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
      IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
      CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
      TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
      SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 */
/*****************************************************************/
/* CHANGES
   Version 1.12  HTTP/1.0 bodies received after a delay would
                 write a zero length buffer through PackListAppend().
		 Added test for empty buffer.

   Version 1.11  Modified URLs requested to eliminate '/./' within
                 HTTP requests.  Netscape Enterprise server, for one,
		 rejects them.

   Version 1.10  Rewrote HTTP handling, other changes for DDJ article.

   Version 1.03  GENMODMASK used when generating mode values.
		 Preserve R lines when generating lists.
		 chmod to allow distribution of mode 04xx files.
		 Need to chmod after transfer.
		 Incorporate umask into chmod call.

   Version 1.02: When could not open destination file for writing,
                 close socket before continuing with next file.

                 Obsolete processing: rmdir() and unlink() return
				 codes checked and warning printed.

                 MODMASK allows x bit to be ignored on MS-DOS file
		 systems.

   Version 1.01: Features added.  Obsolete file removal, file modes,
                 HTTP/1.1 transfers.

   Version 1.00: Introduced.
*/
/*****************************************************************/
/* Setup HTTPSYNC compilation environment.  Handle various operating
   system differences.
 */
 #ifdef NeXT
        #define NO_INC_MALLOC
 #endif
 #ifndef NO_INC_MALLOC
 #include <malloc.h>
 #endif
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <memory.h>
 #include <stddef.h>
 #include <stdlib.h>
 #include <sys/types.h>

/**********************************************************/
/*   WinSock and BSD-style sockets portability            */
/* Macros and conditionals for portable sockets code
 * 1. Write code using readsocket, writesocket, closesocket,
 *    INVALID_SOCKET, SOCKET_ERROR, SOCKET, and INADDR_NONE
 * 2. Always use SocketStartup() and SocketCleanup()
 * 3. Conditional includes and definitions for those macros
 *    allow operation under Windows and BSD-style sockets.
 */
#ifdef WIN32 /* Windows systems */

#include <winsock.h>
#include <io.h>
#define readsocket(a,b,c) recv(a,b,c,0)
#define writesocket(a,b,c) send(a,b,c,0)
/* closesocket() does not need a macro.
 * INVALID_SOCKET, SOCKET_ERROR, SOCKET, and INADDR_NONE
 * are already defined in winsock.h
 */
WSADATA libmibWSAdata;
#define SocketStartup() \
    if (WSAStartup(0x101,&libmibWSAdata)) exit(-1)
#define SocketCleanup() WSACleanup()

#else /* Unix-style systems */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#define readsocket read
#define writesocket write
#define closesocket close
#define SocketStartup()
#define SocketCleanup()
#define INVALID_SOCKET -1
#define SOCKET_ERROR -1
#define SOCKET int
/* define INADDR_NONE if not already */
#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

#endif
/*****************************/
/* Additional macros and environment for portable
 * HTTPsync code
 */
#ifdef WIN32
#include <direct.h>
#include <sys/utime.h>
#define TDIFFLIMIT 2  /* MS-DOS file timestamp resolution */
#define MODMASK 0600  /* When comparing, don't check Execute, group,
			 or other bits */
#define GENMODMASK 0644 /* When generating, use only rw-r-r bits */
#define strncasecmp strnicmp
#define portable_mkdir(d,perm) mkdir(d)
#define FOPEN_WRITE_BINARY "wb"
#else /* Unix-like systems */
#include <sys/ioctl.h>
#include <unistd.h>
#include <utime.h>
#include <errno.h>
#define portable_mkdir(d,perm) mkdir(d,perm)
#define FOPEN_WRITE_BINARY "w"
#define TDIFFLIMIT 1
#define MODMASK 0700  /* Check only that owner modes match */
#define GENMODMASK 0755 /* For generating packing lists. */
#endif

#include <assert.h>
 
#ifdef NeXT
        #include        <objc/hashtable.h>
        #include        <bsd/sys/time.h>
        #define utimbuf myutimbuf
        struct myutimbuf {
                time_t actime;
                time_t modtime;
        };
        #define strdup NXCopyStringBuffer
#endif

/*****************************************************************/
/* -- End setup of compilation environment. There should be no O/S
 *  dependent conditionals below this line.
 */
/*****************************************************************/


/*****************************************************************/

/* Predeclarations. These functions are defined in this file after
   they are referenced.
 */
int ssgmtime(const char *sptr,time_t *time);
int countbl(const char *pStart); /* count consecutive ' ' and '\t' */
int counttoch(const char *pStart,char ch); /* count to \0 or ch */
int counttowh(const char *pStart); /* count to \0 ' ' '\t' '\r' or '\n' */
char *astrn0cpy(char **asz,const char *src,int len);
char *astrn0cat(char **asz,const char *src,int len);
time_t tm_to_time (struct tm *tp);


#define PACKLISTVER 101  /* 100 does not support 'R', 'O' tags */

int bVERBOSE = 0;
int umaskval;
/**************************************************************/


/**************************************************************/
/* HTTPaccess implements a reusable method of making multiple
   requests to an HTTP server.  Tries HTTP/1.1 and falls back
   to HTTP/1.0 after a number of failed HTTP/1.1 attempts.

   For example use, see mainHTTPSYNC(), basically call
   HTTPaccess_Initialize() to initialize a structure with
   the host names of servers, then make multiple requests
   through HTTPaccess_Retrieve()
   
   For advanced usage, customize HTTPaccess_Retrieve() or
   HTTPaccess_SendRequest()

 */

typedef char EXPLAIN; /* For return values that are error
		      explanations including a decimal ASCII
		      number at position 2.  Success is
		      returned as 0 (NULL).
                       */

struct HTTPaccess_s {
    SOCKET s;
    int bPERSIST; /* already a connection open */
    int bHTTP1_1; /* Attempt connection as HTTP/1.1 */
    struct sockaddr_in saddr;
    int bSKIPDNS; /* Set nonzero when saddr is filled in,
                     the lookup is done once */
    int nPort;
    const char *pszProxy;
    const char *pszHost;
    long cbContent; /* Content length, or one of the following */
#define HTTPaccess_CHUNKED -1
#define HTTPaccess_UNTILCLOSE -2
    int cbInbuf;
    char *buf;
    int cbBuf;
    int cAttempt;
};

void HTTPaccess_Initialize(struct HTTPaccess_s *pHTA,
			   const char *pszProxy,
			   const char *pszHost,
			   int nPort,
			   char * buf,int cbBuf)
{ /* buf must be large enough to hold all response
     headers, or error will be returned.
   */
    pHTA->bHTTP1_1 = 1; /* default. Falls back automatically. */
    pHTA->bSKIPDNS = 0;
    pHTA->pszProxy = pszProxy;
    pHTA->pszHost = pszHost;
    pHTA->bPERSIST = 0;
    pHTA->nPort = nPort;
    pHTA->buf = buf;
    pHTA->cbBuf = cbBuf;
} /* HTTPaccess_Initialize */

EXPLAIN *HTTPaccess_OpenConn(struct HTTPaccess_s *pHTA)
{   /* A "helper" function called by HTTPaccess_SendRequest */
    /* Returns error explanation */

    /* Lookup host name */
    const char *pszHost = pHTA->pszHost;
    struct hostent *phostent;

    if (pHTA->bPERSIST) { /* Already a HTTP/1.1 persistent
	                     connection open */
	return 0;
    }

    if (pHTA->pszProxy) {
	pszHost = pHTA->pszProxy;
    }

    if (!pHTA->bSKIPDNS) {
	phostent = gethostbyname(pszHost);

	pHTA->saddr.sin_family = PF_INET;
	pHTA->saddr.sin_port = htons((short) pHTA->nPort);

	if (!phostent) {
	    pHTA->saddr.sin_addr.s_addr = inet_addr(pszHost);
	    if (pHTA->saddr.sin_addr.s_addr == INADDR_NONE) {
		return "E-3-HTTPaccess_OpenConn: could not resolve name";
	    }
	} else {
	    pHTA->saddr.sin_addr.s_addr = 
		*((u_long *) phostent->h_addr);
	}
    }

    pHTA->s = socket(PF_INET,SOCK_STREAM,0);
    if (pHTA->s == INVALID_SOCKET) {
	return "E-1-HTTPaccess_OpenConn: could not get socket";
    }

    /* And connect */
    pHTA->bSKIPDNS = 0;
    if (bVERBOSE) {
	printf("[connect]");
    }
    if (connect(pHTA->s, 
	       (struct sockaddr *) &(pHTA->saddr),
	        sizeof(pHTA->saddr)) == SOCKET_ERROR) {
	closesocket(pHTA->s);
	return "E-2-HTTPaccess_OpenConn: could not connect";
    }
    pHTA->bSKIPDNS = 1;
    return 0;

} /* HTTPaccess_OpenConn */


EXPLAIN *HTTPaccess_GenRequest(struct HTTPaccess_s *pHTA,
			       const char *pszURI)
{/* Generate headers for an HTTP 1.0 or 1.1 request. 
   NOTE: If HTTP/1.1 the caller must append a Content-length
   header. See HTTPaccess_Retrieve() for an example
 */
    if (strlen(pszURI)+strlen(pHTA->pszHost)*2+100 >
	(unsigned) pHTA->cbBuf) {
	return "E-1-HTTPaccess_GenRequest buffer too small";
    }
    if (pHTA->pszProxy) {
	if (pszURI[0] == '/') {
	    sprintf(pHTA->buf,"GET http://%s%s",
		    pHTA->pszHost,pszURI);
	} else {
	    sprintf(pHTA->buf,"GET http://%s/%s",
		    pHTA->pszHost,pszURI);
	}
    } else {
	if (pszURI[0] == '/') {
	    sprintf(pHTA->buf,"GET %s",pszURI);
	} else {
	    sprintf(pHTA->buf,"GET /%s",pszURI);
	}
    }

    if (pHTA->bHTTP1_1) {
	strcat(pHTA->buf," HTTP/1.1\r\n");
	strcat(pHTA->buf,"Host: ");		
	strcat(pHTA->buf,pHTA->pszHost);
	strcat(pHTA->buf,"\r\n");
    } else {
	strcat(pHTA->buf," HTTP/1.0\r\n");
    }

    return 0;
}

EXPLAIN *HTTPaccess_SendRequest(struct HTTPaccess_s *pHTA,
				const char *pszRequest)
{ /* Returns error explanation */
   /* Ensure connection */
    if (HTTPaccess_OpenConn(pHTA)) {
	return "E-1-HTTPaccess_SendRequest: Could not open connection";
    }
    
    /* Send request */
    if (writesocket(pHTA->s,pszRequest,strlen(pszRequest))
	        == SOCKET_ERROR) {
	if (pHTA->bHTTP1_1) {
	    return "E-3-HTTPaccess_SendRequest. HTTP/1.1 write error";
	}
	return "E-2-HTTPaccess_SendRequest: Could not write socket";
	closesocket(pHTA->s);
    }
    return 0;
} /* HTTPaccess_SendRequest */



EXPLAIN * HTTPaccess_ReadHead(struct HTTPaccess_s *pHTA,int *pcbHead)
{/* Read from a socket, appending to a buffer until entire HTTP
  * header is obtained.  (In HTTP the end of header is marked by
  * a blank line.)
  * Returns "E-3-..." if some problem which indicates a retry is
  *                 indicated.
  * Returns "E-1-..." if some problem which indicates to not
  *                 retry.
  */

    int cnt;
    char *ptr;
    
    while(1) { /* Until there is a blank line in the buffer, which
	        * marks the end of headers
		*/
	/* See if a blank line in what was read so far */
	ptr = pHTA->buf;
	if ((pHTA->cbInbuf > 0) &&
	         ((*ptr == '\r')||(*ptr == '\n')) ) {
	    *pcbHead = 0;
	    return 0;
	}
	while(ptr) {
	    ptr = memchr(ptr,'\n',pHTA->buf+pHTA->cbInbuf-ptr);
	    if (ptr && (ptr+1 < pHTA->buf + pHTA->cbInbuf)) {
		ptr++;
		if ((*ptr == '\r')||(*ptr == '\n')) {
		    *pcbHead = ptr - pHTA->buf;
		    return 0;
		}
	    }
	}

	if (pHTA->cbBuf - pHTA->cbInbuf <= 0) {
	    /* Buffer won't hold all headers */
	    return "E-1-HTTPaccess_ReadHead. Excessive response header\n";
	}
	cnt = readsocket(pHTA->s,
			 pHTA->buf+pHTA->cbInbuf,
			 pHTA->cbBuf- pHTA->cbInbuf);
	if (bVERBOSE) {
	    printf("R%d",cnt);
	}
	if (cnt <= 0) {
	    return "E-3-HTTPaccess_ReadHead. read error\n";
	}
	pHTA->cbInbuf += cnt;
	if (strncasecmp(pHTA->buf,"HTTP/1.",7)) {
	    /* No headers (or response not understood */
	    return "E-3-HTTPaccess_ReadHead. Confused headers\n";
	}
    }
} /* HTTPaccess_ReadHead */


EXPLAIN * HTTPaccess_ProcessHead(struct HTTPaccess_s *pHTA,int cbHead)
{
    char *ptr;

    /*******************************************/
    /* Process header part of response. */
    pHTA->cbContent = HTTPaccess_UNTILCLOSE; /* Default: Until socket closes */

    ptr = pHTA->buf;
    ptr += 7;
    if (pHTA->bHTTP1_1 && (*ptr != '0')) { 
	/* HTTP 1.1 connections can be left open */
	pHTA->bPERSIST = 1;
    }
    while(ptr < pHTA->buf + cbHead) {
	ptr = memchr(ptr,'\n',pHTA->buf + cbHead - ptr);
	if (!ptr) {
	    return 0;
	}
	ptr++;
	if (!strncasecmp(ptr,"Transfer-Encoding: ",19)) {
	    /* For HTTP/1.1 only */
	    ptr += 19;
	    ptr += countbl(ptr);
	    if (strncasecmp(ptr,"chunked",7)) {
		/* Don't know how to deal with anything else */
		pHTA->bHTTP1_1 = 0; 
		closesocket(pHTA->s);
		pHTA->bPERSIST = 0;
	        return "E-3-HTTPaccess_ProcessHead. Unsupported Transfer-Encoding\n";
	    }
	    pHTA->cbContent = HTTPaccess_CHUNKED;
	} else if (!strncasecmp(ptr,"Connection: ",12)) {
	    ptr += 12;
	    ptr += countbl(ptr);
	    if (!strncasecmp(ptr,"close",5)) {
		pHTA->bPERSIST = 0;
		if (bVERBOSE) {
		    printf("[C]");
		}
	    }
	} else if (!strncasecmp(ptr,"Content-Length: ",16)) {
	    ptr += 16;
	    ptr += countbl(ptr);
	    pHTA->cbContent = atoi(ptr);
	    if (bVERBOSE) {
		printf("L%ld",pHTA->cbContent);
	    }
	}
	if (*ptr == '\r') {
	    ptr++;
	}
	if (*ptr == '\n') {
	    ptr++;
	    return 0;
	}
    }
    return 0;


} /* HTTPaccess_ProcessHead */




EXPLAIN *HTTPaccess_ReadBody(struct HTTPaccess_s *pHTA,
		 int (*fn)(void *,const char *,int),
		 void *pID) /* passed to fn */
{ /* Read a HTTP body from a buffer and socket.
     Unprocessed data left in the buffer is processed first,
     then data is read from the socket.
     Will stop reading when reach specified length (cbContent) or
     got a chunk-size of 0.
  */
    char *ptr;
    int cbChunk;
    int cnt;

    if (pHTA->cbContent == 0) {
	return 0;
    }
    if (pHTA->cbContent == HTTPaccess_CHUNKED) { /* Chunked Transfer Coding. */
	/*  See RFC 2068 Section 3.6 and 19.4.6 */
	
	cbChunk = 0; /* At start of chunk */
	while(1) {
	    if (cbChunk == 0) { /* Need to read chunk size */
		ptr = memchr(pHTA->buf,'\n',pHTA->cbInbuf);
		if (!ptr) {
		    goto readmore;
		}
		ptr++;
		cbChunk = strtol(pHTA->buf,0,16);
		if (bVERBOSE) {
		    printf("C%d",cbChunk);
		}
		cnt = pHTA->cbInbuf - (ptr - pHTA->buf); /* Remainder */
		if (cbChunk == 0) { /* End of chunks, start of entity headers */
		    if (cnt > 0) {
			/* Move rest of data down to base of buffer */
			memmove(pHTA->buf,ptr,cnt);
			pHTA->cbInbuf = cnt;
		    }
		    if (HTTPaccess_ReadHead(pHTA,&cnt)) {
			return "E-3-HTTPaccess_ReadBody bad chunked encoding reading entity-headers";
		    }

		    /* Discard headers */
		    ptr = pHTA->buf + cnt;
		    if (*ptr == '\r') {
			ptr++;
		    }
		    ptr++;
		    cnt = pHTA->cbInbuf + pHTA->buf - ptr;
		    if (cnt > 0) {
		    	memmove(pHTA->buf,ptr,cnt); /* Move data down to base */
		    }
		    pHTA->cbInbuf = cnt;
		    return 0;
		}
	    } else { /* Still processing a chunk */
		ptr = pHTA->buf;
		cnt = pHTA->cbInbuf;
	    }
	    /* Here with ptr pointing to cnt bytes of unprocessed data */
	    if (cnt >= cbChunk + 2) {
		/* More data than chunk. There must be a CRLF following before
		   processing can be done.
		 */
		if ((*fn)(pID,ptr,cbChunk)) {
		    return "E-2-HTTPaccess_ReadBody callback error";
		}
		ptr += cbChunk;
		cnt = pHTA->cbInbuf - (ptr - pHTA->buf); /* Remainder */
		ptr = memchr(ptr,'\n',cnt);
		if (!ptr) {
		    return "E-3-HTTPaccess_ReadBody bad chunked coding no CRLF at end of chunk";
		}
		ptr++;
		cnt = pHTA->cbInbuf - (ptr - pHTA->buf); /* Remainder */

		pHTA->cbInbuf = cnt;
		memmove(pHTA->buf,ptr,cnt); /* Move data down to base */
		cbChunk = 0;
		continue; /* don't readmore yet.*/
	    } else { /* Write everything we have, possibly reserving 1 byte */
		if (cbChunk - cnt < 1) {
		    cnt--; /* preserve one byte */
		}
		if ((*fn)(pID,ptr,cnt)) {
		    return "E-2-HTTPaccess_ReadBody callback error";
		}
		cbChunk -= cnt;
		pHTA->cbInbuf = ptr + cnt - (pHTA->buf + pHTA->cbInbuf);
		if (pHTA->cbInbuf > 0) {
		    /* Move rest of data down to base of buffer */
		    memmove(pHTA->buf,ptr+cnt,pHTA->cbInbuf);
		}

	    }
readmore:;
	    /* Need to read more */
	    if (pHTA->cbInbuf >= pHTA->cbBuf) { /* No room */
		return "E-3-HTTPaccess_ReadBody bad chunked coding";
	    }
	    cnt = readsocket(pHTA->s,pHTA->buf+pHTA->cbInbuf,pHTA->cbBuf-pHTA->cbInbuf);
	    if (cnt < 0) {
		return "E-1-HTTPaccess_ReadBody socket read error";
	    }
	    pHTA->cbInbuf += cnt;
	}
    } else if (pHTA->cbContent == HTTPaccess_UNTILCLOSE) { /* Read until close */
	while(1) {
	    if (pHTA->cbInbuf > 0) { /* Write what we have */
    		if ((*fn)(pID,pHTA->buf,pHTA->cbInbuf)) {
		    return "E-2-HTTPaccess_ReadBody callback error";
		}
	    }
	    cnt = readsocket(pHTA->s,pHTA->buf,pHTA->cbBuf);
	    if (cnt < 0) {
		return "E-1-HTTPaccess_ReadBody socket read error";
	    }
	    if (cnt == 0) {
		return 0;
	    }
	    pHTA->cbInbuf = cnt;
	}
    } else { /* Read cbContent */
	while(1) {
	    if (pHTA->cbInbuf >= pHTA->cbContent) {
		/* Write partial */
		if ((*fn)(pID,pHTA->buf,pHTA->cbContent)) {
		    return "E-2-HTTPaccess_ReadBody callback error";
		}
		ptr = pHTA->buf + pHTA->cbContent;
		cnt = pHTA->cbInbuf - (ptr - pHTA->buf); /* Remainder */
		if (cnt > 0) {
		    memmove(pHTA->buf,ptr,cnt);
		}
		pHTA->cbInbuf = cnt;
		return 0;
	    } else { /* write everything in the buffer */
		if ((pHTA->cbInbuf > 0) && (*fn)(pID,pHTA->buf,pHTA->cbInbuf)) { /* 5-25-99 */
		    return "E-2-HTTPaccess_ReadBody callback error";
		}
		pHTA->cbContent -= pHTA->cbInbuf;
	    }
	    cnt = pHTA->cbBuf;
	    if (cnt > pHTA->cbContent) {
		cnt = pHTA->cbContent;
	    }
	    cnt = readsocket(pHTA->s,pHTA->buf,cnt);
	    if (cnt < 0) {
		return "E-1-HTTPaccess_ReadBody socket read error";
	    }
	    pHTA->cbInbuf = cnt;
	}
    }
} /* HTTPaccess_ReadBody */



EXPLAIN *HTTPaccess_Retrieve(struct HTTPaccess_s *pHTA,
			        char *pszURI,
				int (*fn)(void *,const char *,int),
				void *pID)
{ /* This is a general purpose function itself, and serves as
     an example of how to create complex HTTP/1.1 requests

     The pID is used as the first argument to the fn, which
     should write the amount of data in the buffer, and return
     0 for success, -1 for not.
   */
char *pszExplain;	
char *ptr;
int errcode;
int cnt;

retryrequest:
    if (pHTA->cAttempt) { /* Retry handling */
	/* A subsequent attempt due to HTTP read or write error
	   which could have been a transient due to network
	   problems, or it could be a poorly implemented proxy
	   or cache.  HTTP/1.1 is more efficient, but this will
	   use HTTP/1.0 after a number of failures of HTTP/1.1
	*/
	if (pHTA->s && pHTA->bPERSIST) {
	    closesocket(pHTA->s);
	}
	pHTA->bPERSIST = 0;
	if (pHTA->cAttempt <= 2) {
	    if (bVERBOSE) {
		printf("Retry...\n");
	    }
	} else if (pHTA->bHTTP1_1) {
	    pHTA->cAttempt = 0;
	    pHTA->bHTTP1_1 = 0;
	    if (bVERBOSE) {
		printf("Retry as HTTP/1.0...\n");
	    }
	} else {
	    return "E-4-HTTPaccess_Retrieve failure after retries";
	}
    }

    pHTA->cAttempt++;

    /* Format the request appropriately for
    the appropriate HTTP version and proxy (if any). */
    if (HTTPaccess_GenRequest(pHTA,pszURI)) {
	fprintf(stderr,"Terminated. Could not format request into buffer\n");
	exit(-1);
    }

    /**** Add additional headers, depending on HTTP version */
    if (pHTA->bHTTP1_1) {
	strcat(pHTA->buf,"Content-Length: 0\r\n");
    }

    /**** Add header separator, and nothing else (empty body) */
    strcat(pHTA->buf,"\r\n");

    if (bVERBOSE) { /* Show entire request headers*/
	fprintf(stderr,"%s",pHTA->buf);
    }
    
    /**** Send the request */
    pszExplain = HTTPaccess_SendRequest(pHTA,pHTA->buf);
    if (pszExplain && (atoi(pszExplain+2) == 3)) {
	goto retryrequest;
    }
    
    /**** Receive response */
    pHTA->cbInbuf = 0;
    pszExplain = HTTPaccess_ReadHead(pHTA,&cnt);
    if (pszExplain && (atoi(pszExplain+2) == 3)) {
	/* Retry if we got E-3 */
	goto retryrequest;
    }
    pszExplain = HTTPaccess_ProcessHead(pHTA,cnt);
    if (pszExplain && (atoi(pszExplain+2) == 3)) { 
	/* Retry if we got E-3 */
	goto retryrequest;
    }
    
    /* Skip forward to response code */
    ptr = pHTA->buf + counttowh(pHTA->buf); 
    ptr += countbl(ptr);

    errcode = atoi(ptr); 

    /**** Handle response */
    if (errcode != 200) {
	ptr = strchr(pHTA->buf,'\n');
	if (!ptr) {
	    ptr = pHTA->buf + strlen(pHTA->buf);
	}
	printf("Error: %s HTTP status %d", pszURI,errcode);
	fwrite(pHTA->buf,1,ptr-pHTA->buf,stdout);
	printf("\n");
	pHTA->bPERSIST = 0;
	closesocket(pHTA->s);
	return "E-1-HTTPaccess_Retrieve Status code is not 200";
    }

    /*******************************************/
    /* Skip past headers and read the rest of the body */
    ptr = pHTA->buf + cnt;
    if (*ptr == '\r') {
	ptr++;
    }
    ptr++;
    cnt = pHTA->buf + pHTA->cbInbuf - ptr;

    if (cnt > 0) {
	memmove(pHTA->buf,ptr,cnt);
    }
    pHTA->cbInbuf = cnt;
    if (HTTPaccess_ReadBody(pHTA,fn,pID)) {
	/* If we get an error here, we have to let the
	caller retry, since some information may have
	been written using the fn and pID.
	*/
	closesocket(pHTA->s);
	pHTA->s = 0;
	pHTA->bPERSIST = 0;
	return "E-3-HTTPaccess_Retrieve error during ReadBody";
    }
    
    /*******************************************/
    /* Finish up this transfer. */
    if (!pHTA->bPERSIST) {
	closesocket(pHTA->s);
	pHTA->s = 0;
    }
    return 0;
} /* HTTPaccess_Retrieve */

/* End of HTTPaccess_ routines */
/**************************************************************/


/**************************************************************/
/* The packing list is implemented as a contiguous block of
   memory which is realloc'ed to add entries.

  Append onto the block with PackListAppend(), this does not
  need to be on line boundaries.

  PackListNext() is used to enumerate the non-comment entries
  in the loaded packing list.

 */
char *pPacklist = 0;
size_t cbPacklist = 0;

int PackListAppend(void *pUnused,const char *buf,int amt)
{
    if (amt <= 0) {
	return -1;
    }
    pPacklist = realloc(pPacklist,cbPacklist + amt + 1);
    if (!pPacklist) {
	fprintf(stderr,"Terminated: Could not realloc %ld\n",
	    (long) cbPacklist + amt + 1);
	exit(-1);
    }
    memcpy(pPacklist+cbPacklist,buf,amt);
    cbPacklist += amt;
    pPacklist[cbPacklist] = 0; /* Always ensure there
                                  is NUL terminator */
    return 0;
} /* PackListAppend */


const char *PackListNext(const char *pLast)
{ /* Iterator for accessing a packing list */
const char *pNext;

    if (!pLast) {
	pNext = pPacklist;
    } else { /* Advance first */
	pNext = strchr(pLast,'\n');
	if (pNext) {
	    pNext++;
	}
    }
    if (!pNext || (pNext >= pPacklist + cbPacklist)) {
	return 0;
    }
    while(1) {
	if ((*pNext == '#')||
	    (*pNext == '\r')||
	    (*pNext == '\n')) { /* Ignore it as a comment */
	    if (!strncasecmp(pNext,"#-#httpsync ",12)) {
		if (atoi(pNext + 12) > PACKLISTVER) {
		    fprintf(stderr,"Terminated: This packing list requires newer version of HTTPsync.\n");
		    exit(-1);
		}
	    }
	    pNext = strchr(pNext,'\n');
	    if (!pNext) {
		return 0;
	    }
	    pNext++;
	    /* loop back around */
	} else if (strstr(pNext,"..")) { /* Disallow */
	    fprintf(stderr,"Terminated. Packing list may not contain \"..\" in file names\n");
	    exit(-1);
	} else if ((*pNext == '.')&&(pNext[1] == '/')) {
	    return pNext; /* A file specifier */
	} else if ((*pNext == 'R')&&(pNext[1] == '/')) {
	    return pNext; /* A redirected URL */
	} else if ((*pNext == 'O')&&(pNext[1] == '.')) {
	    return pNext; /* Obsolete file or directory */
	} else {
	    fprintf(stderr,"Terminated: Packing list syntax. Unrecognized line.\n");
	    exit(-1);
	}
    }
} /* PackListNext */


char *PackListCheckRedirect(char *pszURI)
{
const char *ptr = pPacklist;
const char *ptr2;
char *aszRet = 0;

    while(ptr) {
	if (*ptr == 'R') { /* redirect URL.  There is assumed only one 'R'. */
	    ptr2 = strchr(ptr,'\n');
	    if (!ptr2) {
		ptr2 = ptr + strlen(ptr);
	    }
	    astrn0cpy(&aszRet,ptr+1,ptr2-(ptr+1));
	    if (!aszRet) {
		fprintf(stderr,"Terminated: Could not malloc\n");
		exit(-1);
	    }
	    return aszRet;
	}
	ptr = strchr(ptr,'\n');
	if (ptr) {
	    ptr++;
	}
    }
    return pszURI;	 /* No change */
} /* PackListCheckRedirect */
/**************************************************************/


/**************************************************************/
int mainModReport(int argc,char **argv)
{ /* Generate a version 1.01 packing list from a list of files
     on stdin.  (HTTPsync 1.12 uses version 1.01 packing lists.)

     For each listed file, write out file name, length, GMT
     time stamp, and modes.
     
       Files names must start with './' and not include '..'
    */

char buf[8192];
const char *pNext;
char *pszDest = 0;
int cnt;
struct stat s;
static char *szDays[]= {
	"Sun", "Mon" , "Tue" , "Wed"
                       , "Thu" , "Fri" , "Sat" };
static char *szMonths[] = {"Jan" , "Feb" , "Mar" , "Apr"
                       , "May" , "Jun" , "Jul" , "Aug"
                       , "Sep" , "Oct" , "Nov" , "Dec"};

    /* Read it in */
    while((cnt = read(fileno(stdin),buf,sizeof(buf))) > 0) {
	PackListAppend(0,buf,cnt);
    }

    pNext = PackListNext(0);
    if (pNext && (*pNext != '#')) {
	/* If packing list doesn't start with a comment, then
	   it is probably a plain list.
	   
	   We add HTTPsync a version specifier: 101 for the
	   packing list we are generating.
	*/
        printf("#-#httpsync 101 Packing List for httpsync 1.01-1.12\n");
        printf("# Visit the httpsync home page: http://www.mibsoftware.com/httpsync/\n");
    }
    while (pNext) {
	/* PackListNext() is validating lines and will terminate if
	   there is any trouble with file names, format,
	   unrecognized line types, etc.  The routine here modifies
	   only the lines beginning with '.', the rest are written
	   unchanged.
	 */
	if (*pNext == '.') {
	    /* Find end of file name */
	    /* Make a copy of just the file name */
	    astrn0cpy(&pszDest,pNext,counttowh(pNext));
	    if (!pszDest) {
		fprintf(stderr,"Terminated: Could not realloc %d\n",
		    counttowh(pNext));
		exit(-1);

	    }

	    if (!stat(pszDest,&s)) {
		if (s.st_mode & S_IFDIR) {
		} else {
		    struct tm *tmptr;
		    tmptr = gmtime(&s.st_mtime);
		
		    printf("%s %ld %s, %02d %s %04d %02d:%02d:%02d GMT %o\n",
			pszDest,
			s.st_size,
			szDays[tmptr->tm_wday],
			tmptr->tm_mday,
			szMonths[tmptr->tm_mon],
			tmptr->tm_year+1900,
			tmptr->tm_hour,
			tmptr->tm_min,
			tmptr->tm_sec,
			s.st_mode & GENMODMASK);
		}
	    } else { /* can't stat.  must be obsolete. */
		printf("O%s\n", pszDest);
	    }
	} else { /* write it unchanged */
	    fwrite(pNext,1,counttoch(pNext,'\n'),stdout);
	    fwrite("\n",1,1,stdout);
	}
	pNext = PackListNext(pNext);
    }	
    if (pszDest) {
	free(pszDest);
    }
    return 0;
} /* mainModReport */
/**************************************************************/



/**************************************************************/
/* Arguments whic are processed from command line */
char *pszURI;
char *pszProxy = 0;
char *pszHost;


int mainHTTPSYNC(char *pszProxy, char *pszHost, int nPort);

int main(int argc,char **argv)
{
    char *ptr;
    int argind = 1;
    int nPort = 80; /* Default */

    if (argc < 2) {
	fprintf(stderr,"httpsync V 1.12. Copyright 1999 Mib Software www.mibsoftware.com.\n");
	fprintf(stderr,"Usage (get update): httpsync [-x proxy-host] @http://host/URI\n");
	fprintf(stderr,"Usage (create list): httpsync -m <files.lst >packing.lst\n");

	exit(-1);
    }
    if (!strcmp(argv[1],"-m")) {
	return mainModReport(argc,argv);
    }
    while((argind < argc) && (argv[argind][0] == '-')) { 
	switch(argv[argind][1]) {
	case 'x':/* Use a proxy */
	    if (!argv[argind][2]) {
		argind++;
		pszProxy = argv[argind];
	    } else {
		pszProxy = argv[argind]+2;
	    }
	    argind++;
	    break;
	case 'v': /* Mostly useful for debugging.  Not documented. */
	    bVERBOSE = 1;
	    argind += 1;
	    break;
	default:
	    fprintf(stderr,"Terminated: unknown option '%c'\n",argv[argind][1]);
	    exit(-1);
	    break;
	}
    }
    /* Get current umask */
    umaskval = umask(0);
    umask(umaskval);

    if (strncasecmp(argv[argind],"@http://",8)) {
	fprintf(stderr,"Terminated: URL syntax\n");
	exit(-1);
    }
    /* Strip out the host */
    pszHost = argv[argind]+8;
    if (!(ptr = strchr(pszHost,'/'))) {
	fprintf(stderr,"Terminated: URL syntax. No URI specified\n");
	exit(-1);
    }
    *ptr = 0;
    pszURI = ptr+1;
    if ((ptr = strchr(pszHost,':'))) {
	*ptr = 0;
	nPort = atoi(ptr+1);
    }
    if (pszProxy) {
	if ((ptr = strchr(pszProxy,':'))) {
	    *ptr = 0;
	    nPort = atoi(ptr+1);
	}
    }
    /* End of command line parameter processing */
    /********************************************/

    return mainHTTPSYNC(pszProxy,pszHost,nPort);
} /* main */
/**************************************************************/



/**************************************************************/
int fnWriteFile(void *v,const char *p,int cb)
{ /* Helper function */
    FILE *f = (FILE *) v;
    if (write(fileno(f),p,cb) >= 0) {
	return 0;
    }
    return -1;
}

/**************************************************************/

const char *TransferDecide(const char *pszDest,
			   const char *ptr,
			   time_t *ptLastmod,
			   int *pbvMode); /* helper */
void EnsureDirectoryForFile(const char *pszDest); /* helper */

int mainHTTPSYNC(char *pszProxy, char *pszHost, int nPort)
{
    /* Main loop */
    char buf[8192];
    
    time_t tLastmod;
    long cbLastmod = 0;
    int cUnmodified = 0;
    char *pszDest = 0;
    const char *pNext;
    FILE *fDEST;
    const char *ptr;
    const char *pszTransferReason ="";
    int bvMode;
    struct HTTPaccess_s theHTA;
    struct HTTPaccess_s *pHTA = &theHTA;
    struct stat s;
    int cURIprefix;

    SocketStartup();
    pszDest = "";

    HTTPaccess_Initialize(pHTA,pszProxy,pszHost,nPort,buf,sizeof(buf));

/* Retrieve and store the packing list in memory. The packing list
   may specify a different file name than the URL, so the packing list
   is written to a file after retreival is complete.
 */
    printf("GET %s...", pszURI);
    pHTA->cAttempt = 0;
    while(1) {
	fflush(stdout);
	/* Initialize for read */
	cbPacklist = 0;

	ptr = HTTPaccess_Retrieve(pHTA,pszURI,PackListAppend,0);
	
	if (!ptr) break; /* success */
	
	if (atoi(ptr+2)!=3) {
	    fprintf(stderr,"Terminated.  can't get packing list\n");
	    exit(-1);
	}
	 /* retry */
    }
    printf("OK\n");
    fflush(stdout);
    pszURI = PackListCheckRedirect(pszURI);

    /* Copy the URI into an allocated buffer to allow
       strcat of the filenames later.
    */
    ptr = pszURI;
    pszURI = 0;
    astrn0cpy(&pszURI,ptr,strlen(ptr));

    /* Save the file */
    ptr = strrchr(pszURI,'/');
   
    if (!ptr) {
	fprintf(stderr,"Terminated: No '/' in packing list URI\n");
	exit(-1);
    }
    ptr++;
    fDEST = fopen(ptr,FOPEN_WRITE_BINARY);
    if (!fDEST) {
	fprintf(stderr,"Terminated: can't save '%s'.  (Packing list missing 'R'?)\n",ptr);
	exit(-1);
    }
    if (fwrite(pPacklist,1,cbPacklist,fDEST) != cbPacklist) {
	fprintf(stderr,"Terminated: can't save '%s'. write error\n",ptr);
	exit(-1);
    }
    fclose(fDEST);

    cURIprefix = ptr - pszURI; /* pszURI is a prefix.  Packing list filenames will be
	         appended at this point */

    pNext = PackListNext(0);
    while(pNext) { /* For each entry in the packing list */
	fDEST = 0;
	tLastmod = 0;
	cbLastmod = -1;
	bvMode = 0;

	pszURI[cURIprefix] = 0;
	/* 5-25-99 Modified to eliminate /./ requests which 
	 * are rejected by Netscape enterprise server 
	 */
	astrn0cat(&pszURI,pNext+2,counttowh(pNext+2)); 

	pszDest = pszURI + cURIprefix;
	ptr = pNext + counttowh(pNext);
	ptr += countbl(ptr);

	if (*pNext == 'R') {
	  /* Loop back around 'R' was already processed */
	} else if (*pNext == 'O') {
	    /* Remove obsolete file */
	    if (!stat(pszDest+1,&s)) { /* It exists */
		if (s.st_mode & S_IFDIR) {
		    printf("Removing obsolete directory: %s\n",
			  pszDest+1);
		    if (rmdir(pszDest+1)) {
		      printf("...Warning: Directory not removed: %s\n",
			  strerror(errno));
		    }
		} else {
		    printf("Removing obsolete file: %s\n",pszDest+1);
		    if (unlink(pszDest+1)) {
		      printf("...Warning: File not removed: %s\n",strerror(errno));
		    }
		}
	    }
	} else if (*pNext == '.') {
	    bvMode = 0777 & ~umaskval; /* default */
	    pszTransferReason = TransferDecide(pszDest,ptr,&tLastmod,&bvMode);
	    if (!pszTransferReason) { /* Don't transfer.  It matches */
		cUnmodified++;
	    } else if (!strcmp(pszTransferReason,"M")) { /* Just a mode change */
		printf("%s %s\n",pszTransferReason,pszDest);
		chmod(pszDest,bvMode & ~umaskval); /* Attempt it */
	    } else {
		/* Otherwise: transfer is required */
		printf("%s %s...", pszTransferReason,pszDest);
		fflush(stdout);
		pHTA->cAttempt = 0;
		while(1) {
		    fDEST = fopen(pszDest,FOPEN_WRITE_BINARY);
		    if (!fDEST) {
			/* Fix common reasons file could not be opened */
			if ((bvMode & 0600) == 0400) { /* Exists as 04xx read only */
			    chmod(pszDest,bvMode | 0200);
			}
			/* Ensure the directory */
			EnsureDirectoryForFile(pszDest);
			fDEST = fopen(pszDest,FOPEN_WRITE_BINARY);
		    }
		    if (!fDEST) {
			fprintf(stderr,
			    "Terminated. Could not open %s for writing\n",
			    pszDest);
			exit(-1);
		    }
		    ptr = HTTPaccess_Retrieve(pHTA, pszURI,
			         fnWriteFile,
			         fDEST); 
		    fclose(fDEST);
		    if (!ptr) break; /* success */
		    if (atoi(ptr+2)!=3) {
			fprintf(stderr,"Terminated. Could not transfer %s\n",pszDest);
			exit(-1);
		    }
		    /* retry */
		}

		/* Fix timestamp and mode bits */
		if (tLastmod) {
		    struct utimbuf u;
		    u.actime = time(NULL);
		    u.modtime = tLastmod;
		    utime(pszDest,&u);
		}
		chmod(pszDest,bvMode & ~umaskval);
		printf("OK\n");
	    }
	}
	
	/* Advance */
	pNext = PackListNext(pNext);
    }
    
    printf("Skipped %d unchanged files\n",cUnmodified);
    if (pHTA->s && pHTA->bPERSIST) {
	closesocket(pHTA->s);
    }
    SocketCleanup();
    return 0;
} /* mainHTTPsync */
/**************************************************************/


/**************************************************************/
/******** Supporting Functions ********************************/
/**************************************************************/
const char *TransferDecide(const char *pszDest,
			   const char *ptr,
			   time_t *ptLastmod,
			   int *pbvMode)
{ /* Make the decision to transfer or not transfer a file.
  
     pszDest is a file name.
     ptr is the pointer to the packing list information:
     length, time stamp, etc

    return 0 if unmodified, otherwise a short string which
    notes one reason the local doesn't match the remote.
  */

    int cbLastmod = atoi(ptr); /* Length */
    struct stat s;

    ptr = strchr(ptr,' '); /* Skip ahead */
    if (!ptr) {
      /* Packing list syntax */
      fprintf(stderr,"Terminated: Packing list syntax. File status information missing\n");
      exit(-1);
    }

    ptr += countbl(ptr);
    if (strchr(ptr,',')) {
      ptr = strchr(ptr,',')+1;
      ptr += countbl(ptr);
    }
    ptr += ssgmtime(ptr,ptLastmod);
    ptr += countbl(ptr);
    if (*ptr <= ' ') {
      fprintf(stderr,"Terminated: Packing list syntax. File status information missing\n");
      exit(-1);
    }

    /* Permissions */
    *pbvMode = strtol(ptr,(char **) &ptr,8);
    *pbvMode &= 0777;

    if (!stat(pszDest,&s)) {
	if (s.st_mtime < *ptLastmod) {
	  return("D.lt");
	} else if(s.st_mtime - *ptLastmod >= TDIFFLIMIT) {
	  return("D.gt");
	} else if (s.st_size != cbLastmod) {
	  return("S.ne");
	} else {
	  /* Everything else matches.  Check perms */
	  if (*pbvMode && 
	      ((s.st_mode & 0700) != (*pbvMode & MODMASK))) {
	      return("M");
	  }
	  if (bVERBOSE) {
	      printf("%s Unmodified from packing list\n",
		     pszDest);
	  }
	  return 0;
	}
    } else {
	return("N");
    }

} /* TransferDecide */
/**************************************************************/


/**************************************************************/
void EnsureDirectoryForFile(const char *pszDest)
{ /* Create each directory in a chain.  Ignore errors, since it
     is likely that we will run mkdir() on directories which
     already exist.
   */
char *work = 0;
char *ptr;
    astrn0cpy(&work,pszDest,strlen(pszDest));
    if (!work) {
	fprintf(stderr,"Terminated. Can't realloc\n");
	exit(-1);
    }
    ptr = work;
    while((ptr = strchr(ptr,'/'))) {
	*ptr = 0;
	portable_mkdir(work,0755 & ~umaskval);

	*ptr++ = '/';
    }
    free(work);
} /* EnsureDirectoryForFile */
/**************************************************************/

/** ssgmtime()
    For HTTPsync, it is known that the date string form will always
    be
        DD 3-letter-Month YYYY HH:MM:SS GMT
    
    The day of week was skipped before calling here.

    This loads the values into a struct tm, and then converts to
    a time_t.  mktime() is not used since we can't figure out
    timezone in a portable way that is reliable across daylight
    savings, etc.

 **/
int ssgmtime(const char *sptr,time_t *time)
{ 
const char *ptr;
struct tm tmbuf;

static char *szMonths[] = {"Jan" , "Feb" , "Mar" , "Apr"
                       , "May" , "Jun" , "Jul" , "Aug"
                       , "Sep" , "Oct" , "Nov" , "Dec"};

  ptr = sptr;
  /* Parse for the date. */
  tmbuf.tm_mday = strtol(ptr,(char **) &ptr,10);
  ptr += countbl(ptr);
  
  for(tmbuf.tm_mon = 0;tmbuf.tm_mon < 12;tmbuf.tm_mon++) {
      if (!strncasecmp(szMonths[tmbuf.tm_mon],ptr,3)) {
	  break;
      }
  }
  if (tmbuf.tm_mon >= 12) {
      return 0;
  }
  ptr += 3;
  ptr += countbl(ptr);

  tmbuf.tm_year = strtol(ptr,(char **) &ptr,10) - 1900;

  tmbuf.tm_hour = strtol(ptr,(char **) &ptr,10);
  if (*ptr != ':') {
      return 0;
  }
  tmbuf.tm_min = strtol(ptr+1,(char **) &ptr,10);
  if (*ptr != ':') {
      return 0;
  }
  tmbuf.tm_sec = strtol(ptr+1,(char **) &ptr,10);
  ptr += countbl(ptr);
  if ((ptr[0]=='G')&&(ptr[1]=='M')&&(ptr[2]=='T')&&isspace(ptr[3])) {
      ptr += 3;
  } else {
      return 0;
  }
  *time = tm_to_time(&tmbuf);
  return ptr - sptr;
} /* ssgmtime */
/*********************************************************/


/*********************************************************/
/* Automatic allocating string functions */
char *astrn0cpy(char **ppasz,const char *src,int len)
{ /* allocate and strncpy, always store '\0' */
    *ppasz = realloc(*ppasz,len+1);
    if (!*ppasz) {
	return 0;
    }
    strncpy(*ppasz,src,len);
    (*ppasz)[len] = 0;
    return *ppasz;
} /* astrn0cpy */

char *astrn0cat(char **ppasz,const char *src,int len)
{ /* reallocate and strncat, always store '\0' */
int cBefore = 0;
    if (*ppasz) {
        cBefore = strlen(*ppasz);
    }
    *ppasz = realloc(*ppasz,len+1+cBefore);
    if (!*ppasz) {
	return 0;
    }
    strncat(*ppasz+cBefore,src,len);
    (*ppasz)[cBefore + len] = 0;
    return *ppasz;
} /* astrn0cat */
/*********************************************************/


/*********************************************************/
/* Simple consecutive character counting functions  */
int counttowh(const char *pStart)
{
const unsigned char *ptr;
    ptr = (const unsigned char *) pStart;
    while(*ptr && (*ptr > ' ')) {
	ptr++;
    }
    return (char *) ptr - pStart;
} /* counttowh */

int countbl(const char *pStart)
{
const char *ptr;
    ptr = pStart;
    while((*ptr == ' ')||(*ptr == '\t')) {
	ptr++;
    }
    return (int) (ptr - pStart);
} /* countbl */

int counttoch(const char *pStart,char ch)
{
const char *ptr = pStart;
    while(*ptr && (*ptr != ch)) {
	    ptr++;
    }
    return ptr - pStart;
} /* counttoch */
/**************************************************************/


/**************************************************************/
/* This tm_to_time software obtained from comp.sources.unix archive
   http://sources.isc.org/dirlist.perl?dir=devel/func/time/&tarball=tm_to_time/tm_to_time/tm_to_time.c

   The timezone correction is not needed for HTTPsync, and is "#if 0" removed.
 */
/**************************************************************/
#include <sys/types.h>

/* Return 1 if `y' is a leap year, 0 otherwise.
 */

static int leap (int y) {
    y += 1900;
    if (y % 400 == 0)
	return (1);
    if (y % 100 == 0)
	return (0);
    return (y % 4 == 0);
}

/* Return the number of days between Jan 1, 1970 and the given
 * broken-down time.
 */

static int ndays (struct tm *p)  {
    register n = p->tm_mday;
    register m, y;
    register char *md = "\37\34\37\36\37\36\37\37\36\37\36\37";

    for (y = 70; y < p->tm_year; ++y) {
	n += 365;
	if (leap (y)) ++n;
    }
    for (m = 0; m < p->tm_mon; ++m)
	n += md[m] + (m == 1 && leap (y));
    return (n);
}

/* Convert a broken-down time (such as returned by localtime())
 * back into a `time_t'.
 */

time_t tm_to_time (struct tm *tp)  {
#if 0 /* Don't need it in httpsync */
    register int m1, m2;
#endif
    time_t t;
#if 0 /* Don't need it in httpsync */
    struct tm otm;
#endif

    t = (ndays (tp) - 1) * 86400L + tp->tm_hour * 3600L
	+ tp->tm_min * 60 + tp->tm_sec;
#if 0 /* Don't need it in httpsync */
    /*
     * Now the hard part -- correct for the time zone:
     */
    otm = *tp;
    tp = localtime (&t);
    m1 = tp->tm_hour * 60 + tp->tm_min;
    m2 = otm.tm_hour * 60 + otm.tm_min;
    t -= ((m1 - m2 + 720 + 1440) % 1440 - 720) * 60L;
#endif

    return (t);
}

/**************************************************************/
/* This strtol software obtained from FreeBSD cvsweb
 */
/**************************************************************/

/*-
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)strtol.c	8.1 (Berkeley) 6/4/93";
#endif /* LIBC_SCCS and not lint */

#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>


/*
 * Convert a string to a long integer.
 *
 * Ignores `locale' stuff.  Assumes that the upper and lower case
 * alphabets and digits are each contiguous.
 */
long
strtol(nptr, endptr, base)
	const char *nptr;
	char **endptr;
	register int base;
{
	register const char *s = nptr;
	register unsigned long acc;
	register unsigned char c;
	register unsigned long cutoff;
	register int neg = 0, any, cutlim;

	/*
	 * Skip white space and pick up leading +/- sign if any.
	 * If base is 0, allow 0x for hex and 0 for octal, else
	 * assume decimal; if base is already 16, allow 0x.
	 */
	do {
		c = *s++;
	} while (isspace(c));
	if (c == '-') {
		neg = 1;
		c = *s++;
	} else if (c == '+')
		c = *s++;
	if ((base == 0 || base == 16) &&
	    c == '0' && (*s == 'x' || *s == 'X')) {
		c = s[1];
		s += 2;
		base = 16;
	}
	if (base == 0)
		base = c == '0' ? 8 : 10;

	/*
	 * Compute the cutoff value between legal numbers and illegal
	 * numbers.  That is the largest legal value, divided by the
	 * base.  An input number that is greater than this value, if
	 * followed by a legal input character, is too big.  One that
	 * is equal to this value may be valid or not; the limit
	 * between valid and invalid numbers is then based on the last
	 * digit.  For instance, if the range for longs is
	 * [-2147483648..2147483647] and the input base is 10,
	 * cutoff will be set to 214748364 and cutlim to either
	 * 7 (neg==0) or 8 (neg==1), meaning that if we have accumulated
	 * a value > 214748364, or equal but the next digit is > 7 (or 8),
	 * the number is too big, and we will return a range error.
	 *
	 * Set any if any `digits' consumed; make it negative to indicate
	 * overflow.
	 */
	cutoff = neg ? -(unsigned long)LONG_MIN : LONG_MAX;
	cutlim = cutoff % (unsigned long)base;
	cutoff /= (unsigned long)base;
	for (acc = 0, any = 0;; c = *s++) {
		if (!isascii(c))
			break;
		if (isdigit(c))
			c -= '0';
		else if (isalpha(c))
			c -= isupper(c) ? 'A' - 10 : 'a' - 10;
		else
			break;
		if (c >= base)
			break;
		if (any < 0 || acc > cutoff || (acc == cutoff && c > cutlim))
			any = -1;
		else {
			any = 1;
			acc *= base;
			acc += c;
		}
	}
	if (any < 0) {
		acc = neg ? LONG_MIN : LONG_MAX;
		errno = ERANGE;
	} else if (neg)
		acc = -acc;
	if (endptr != 0)
		*endptr = (char *)(any ? s - 1 : nptr);
	return (acc);
}

