/* com/syncbuilder/app/SyncAppSkeleton.java:
*
* Copyright (C) 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: SyncAppSkeleton.java,v 1.4 1999/02/26 22:55:24 christ Exp $ */

package com.syncbuilder.app;


import com.syncbuilder.util.getopt.*;
import com.syncbuilder.sync.*;


/**
 * This code can be used as a foundation for all stand-alone synchronization
 * utilities. It can parse command-line parameters, or it can instead ask
 * the parameters through a GUI. Once the parameters have been retrieved,
 * it will establish a connection to the device and invoke the serviceDevice()
 * method, which you will have to provide. The connection will be closed 
 * either after your code has finished its job, or after it has failed 
 * because of an Exception.
 *
 * @author Tilo Christ
 */
public class SyncAppSkeleton
	extends com.syncbuilder.app.BasicAppSkeleton
{
	/**
	* This method needs to be overwritten by your application. 
	* It will be invoked by the skeleton after all options have been
	* parsed, and after a connection to the Palm device has been
	* established.<br>
	* You should use the supplied Link, and you should not try to
	* close the connection yourself, since this will be done for you
	* by the skeleton.
	*
	* @param link the open Link between the host and the Palm device
	* @exception Exception anything can go wrong
	*/
	public void synchronizeDevice(Link link)
		throws Exception
	{
	}


	protected void addFixedOptions()	
	{
		options.addElement(new FlagOpt("jhotsync", "Use experimental JHotSync protocol stack"));
		options.addElement(new StringOpt("port", "Port to use (\".\" means Network HotSync)", false));
	}
	

	public void appMain()
	{
		String port = getStringOpt("port");
		boolean jhotsync = getFlagOpt("jhotsync");
		
		if (port.equals("."))
		{
			try
			{
				Socket.setSocketFactory( new NetworkHSSocketImplFactory() );
			}
			catch(java.io.IOException ioe)
			{
				displayMessage("Initialization of Network HotSync failed!", "abort");
				return;
			}
			
			System.err.println("Initiate Network HotSync now."); 
		}
		else
		{
			if (jhotsync)
			{
				// I am using this way of instantiating the SocketImpl, so that my code does not depend
				// on JHotSync for compilation. 
				try
				{
					Socket.setSocketFactory( (SocketImplFactory)(Class.forName("com.syncbuilderx.sync.JHotSyncSocketImplFactory").newInstance()) );
				}
				catch(Exception e)
				{
					displayMessage("Error setting up JHotSync", "abort");
					return;
				}
			}
			System.err.println("Please hit the HotSync button now."); 
		}
		

		Dlp dlp = null;
		try
		{
			ServerSocket ssock = new ServerSocket(port);

			Link link = ssock.accept();
			dlp = link.getDlp();

			synchronizeDevice(link);
		}
		catch(Exception e)
		{
			displayMessage("Synchronization failed:\n" + e.getMessage(), "abort" );
		}
		finally
		{
			try
			{
				dlp.endHotSync();
			}
			catch (Exception e)
			{
				displayMessage("Connection could not be properly closed", "abort" );		
			}
		}
	}
}
