/* com/syncbuilder/device/StorageInfo.java:
*
* Copyright (C) 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: StorageInfo.java,v 1.1 1999/01/02 11:12:47 christ Exp $ */

package com.syncbuilder.device;

/**
* Information about the Palm device's storage system.
* <br>Currently available Palm devices only have one built-in storage card which contains between
* 128KB and 2MB of RAM, and between 512KB and 2MB of ROM.
* This may change in future devices.
* <br>Palm devices running Palm OS 2.0+ are capable of reporting the number of databases which are
* contained in their storage system.
*
* <br>The information about the actual cards is stored in objects of the CardInfo class.
*
* @see com.syncbuilder.device.CardInfo
*
* @author Tilo Christ
*/
public class StorageInfo 
{
	public StorageInfo()
	{
		_storage_info_ex = false;
	}

	public StorageInfo(int ram_db_count, int rom_db_count)
	{
		this._ram_db_count = ram_db_count;
		this._rom_db_count = rom_db_count;

		_storage_info_ex = true;;
	}

	public int getCardInfoCount()
	{
		return _card_info.size();
	}

	public CardInfo getCardInfo(int index)
	{
		return (CardInfo)(_card_info.elementAt(index));
	}

	/**
	* This is a convenience method to retrieve the CardInfo object of the first built-in card.
	*/
	public CardInfo getCardInfo()
	{
		return (CardInfo)(_card_info.elementAt(CardInfo.DEFAULTCARD));
	}

	public void addCardInfo(CardInfo card_info)
	{
		_card_info.addElement(card_info);
	}

	/**
	* Does this StorageInfo object contain information about the number of databases?
	*/
	public boolean isStorageInfoEx()
	{
		return _storage_info_ex;
	}

	public int getRAMDBCount()
	{
		return _ram_db_count;
	}

	public int getROMDBCount()
	{
		return _rom_db_count;
	}


	public String toString()
	{
		StringBuffer result = new StringBuffer();
		
		result.append("<com.syncbuilder.device.StorageInfo: Number of cards: " + _card_info.size() + "\n");
	
		for (int i = 0; i < _card_info.size(); i++)
		{
			result.append((_card_info.elementAt(i)).toString());
			result.append("\n");
		}

		if (_storage_info_ex)
		{
			result.append("\nRAM Database count: " + _ram_db_count + " ROM Database count: " + _rom_db_count);
		}

		result.append(">");
		
		return result.toString();
	}


	private boolean _storage_info_ex;
	private java.util.Vector _card_info = new java.util.Vector();

	private int _ram_db_count;
	private int _rom_db_count;
}
