/* com/syncbuilder/device/SystemFeature.java:
*
* Copyright (C) 1999, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: SystemFeature.java,v 1.4 1999/01/16 13:22:05 christ Exp $ */

package com.syncbuilder.device;

import com.syncbuilder.storage.Char4;

/**
* Various constants and helper-methods to deal with the System Feature.
* The System Feature contains information about the device's OS version.
* The same information can be found through the SysInfo class as well.
*
* @author Tilo Christ
* @see com.syncbuilder.device.SysInfo
*/
public class SystemFeature
{
	/** The Creator ID of the System Feature */
	public final static Char4 CREATOR = new Char4("psys");

	/** The number of the System Feature */
	public final static int NUMBER = 1;
	

	/**
	* Get the major version number of the Palm device's system ROM. Upgrade patches are taken into
	* account by this method.
	*/
	public static int getRomVersionMajor(int value)
	{
		return (value & 0xff000000) >> 24;
	}

	/**
	* Get the minor version number of the Palm device's system ROM. Upgrade patches are taken into
	* account by this method.
	*/
	public static int getRomVersionMinor(int value)
	{
		return (value & 0x00ff0000) >> 16;
	}

	/**
	* Get the version number of the Palm device's system ROM formatted as a displayable string (e.g. 2.05).
	* Upgrade patches are taken into account by this method.
	*/
	public static String getRomVersionString(int value)
	{
		int major     = (value & 0xff000000) >> 24;
		int minor     = (value & 0x00ff0000) >> 16;
		String minor_string = "0" + Integer.toString(minor);
//		int sub_minor = (_rom_sw_version & 0x0000ffff);

		return major + "." + minor_string.substring(minor_string.length() - 2, minor_string.length());
	}

	/** This class shall not be instantiated */
	private SystemFeature()
	{
	}
}
