/* com/syncbuilder/internal/sync/NetworkHSSocketImpl.java:
*
* Copyright (C) 1998, 1999, Tilo Christ
* Copyright (C) 1998, Robert Crawford
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: NetworkHSSocketImpl.java,v 1.12 1999/03/18 22:37:07 christ Exp $ */

package com.syncbuilder.internal.sync;

import java.io.*;

/**
* This is an implementation of SocketImpl that provides the functionality for
* TCP/IP-based Network HotSync(R).<p>
* This implementation is designed to be used in an environment where
* many Palm devices connect simultaneously with the same server.
*
* @author Tilo Christ, Robert Crawford
*/
public class NetworkHSSocketImpl
	extends com.syncbuilder.sync.SocketImpl
{
// =============================== Socket ==============================================================

	public boolean isOpen()
	{
		return _is_open;
	}

	public int getVersion() 
		throws IOException
	{
		return _version;
	}
	
	/**
	* Read a specified number of bytes into the specified array.
	*/
	public int read(byte[] data, int len) 
		throws IOException
	{
		if (_DEBUG)
			System.err.println("Reading up to " + len + " bytes.");
			
		byte[] buf = new byte[6];
		_in.read(buf);

		if (_DEBUG)
			System.err.println("Header:\n" + com.syncbuilder.util.Util.hexPrint(buf));

		int rlen = com.syncbuilder.util.PackUtils.getLong(buf, 2);

		if (_DEBUG)
			System.err.println("rlen = " + rlen);

		int read = 0;
		while (read < rlen) 
		{
			int n = _in.read(data, read, rlen - read);
			read += n;
		}

		_xid++;
	
		_xid &= 0xff;

		if (_DEBUG)
			System.err.println("Read " + read + " bytes:\n" + com.syncbuilder.util.Util.hexPrint(data, read));

		return read;
	}
	
	/**
	* Write a specified number of bytes from the specified array.
	*/
	public int write(byte[] data, int len) 
		throws IOException
	{
		if (_DEBUG)
			System.err.println("Sending " + len + " bytes:\n" + com.syncbuilder.util.Util.hexPrint(data, len));

		byte[] buf = new byte[6];
		buf[0] = 1;
		buf[1] = (byte)_xid;
		com.syncbuilder.util.PackUtils.setLong(buf, 2, len);
		
		_out.write(buf);
		
		_out.write(data, 0, len);
		
		return 0;
	}


// ============================= ServerSocket ==============================================================

	/** Switch the SocketImpl into ServerSocket-mode */
	public void create()
		throws IOException
	{
	}

    public void accept(com.syncbuilder.sync.SocketImpl impl, int timeout) 
		throws IOException
	{
		if (!(impl instanceof NetworkHSSocketImpl))
			throw new IllegalArgumentException("impl is not a NetworkHSSocketImpl!");
		
		NetworkHSSocketImpl nwhs_impl = (NetworkHSSocketImpl)impl;

		_servsock.setSoTimeout( timeout );
		nwhs_impl._socket = _servsock.accept();  // accept will throw an exception and thus end this method
												 // if the timeout has been reached.

		nwhs_impl._in  = nwhs_impl._socket.getInputStream();
		nwhs_impl._out = nwhs_impl._socket.getOutputStream();
		
		_sendNHSInitSequence(nwhs_impl);
		
		nwhs_impl._is_open = true;
	}

    public void accept(com.syncbuilder.sync.SocketImpl impl) 
		throws IOException
	{
		accept(impl, 0);
	}

	/**
	* Bind the server-socket to the network-device
	*
	* @param device this parameter will be cheerfully ignored
	*/
	public void bind(String device) 
		throws IOException
	{
	}

	/**
	* Make the ServerSocket listen to the network
	*/
    public void listen(int backlog) 
		throws IOException
	{
		_servsock = new java.net.ServerSocket(14238, backlog);
	}


	/** Return the maximum number of simultaneous connections which can be handled by this kind of socket */
	public int getMaxConnections()
	{
		return Integer.MAX_VALUE;   // No limits :-)
	}


// ============================= Socket & ServerSocket ====================================================

	public void close() 
		throws IOException, com.syncbuilder.sync.DlpException
	{
		if (_socket != null)
		{
			_socket.close();
			_is_open = false;
		}
			
		if (_servsock != null)
		{
			_servsock.close();
		}
	}
	

	private static void _sendNHSInitSequence(NetworkHSSocketImpl impl)
		throws IOException
	{
	    byte msg1[] = {
			(byte)0x12, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x24, (byte)0xff, (byte)0xff,
			(byte)0xff, (byte)0xff, (byte)0x3c, (byte)0x00, (byte)0x3c, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xc0, (byte)0xa8,
			(byte)0xa5, (byte)0x1f, (byte)0x04, (byte)0x27, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
			(byte)0x00, (byte)0x00 };

		byte msg2[] = {
			(byte)0x13, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0xff, (byte)0xff,
			(byte)0xff, (byte)0xff, (byte)0x00, (byte)0x3c, (byte)0x00, (byte)0x3c, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x00,
			(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00 };

		byte[] buffer = new byte[200];

		if (_DEBUG)
			System.err.println("Sending InitSequence");

		InputStream in = impl.getInputStream();
		OutputStream out = impl.getOutputStream();

		in.read(buffer, 0, 22);
  		out.write(msg1, 0, msg1.length);
  		in.read(buffer, 0, 50);
  		out.write(msg2, 0, msg2.length);
  		in.read(buffer, 0, 8);

		if (_DEBUG)
			System.err.println("Sent InitSequence");

		if (_DEBUG)
			System.err.println("Probing DL-version...");
		
		try
		{
			com.syncbuilder.internal.sync.DlpImpl dlp_impl = new com.syncbuilder.internal.sync.DlpImpl(in, out);
			com.syncbuilder.device.SysInfo sysinfo = dlp_impl.readSysInfoVer();

			impl._version = sysinfo.getDLPVersion();

			if (_DEBUG)
				System.err.println("Probed DL-version is: " + impl._version);
		}
		catch(Exception e)
		{
			if (_DEBUG)
				e.printStackTrace();
			impl._version = com.syncbuilder.sync.Dlp.DL_1_1;

			if (_DEBUG)
				System.err.println("Probing failed. Using DL 1.1");
		}
	}


	protected boolean _is_open = false;
	
	protected java.net.Socket _socket = null;
	private java.net.ServerSocket _servsock = null;

	private java.io.InputStream  _in  = null;
	private java.io.OutputStream _out = null;

	private int _xid = 0;

	private int _version;

	private final static boolean _DEBUG = false;
}
