/* com/syncbuilder/storage/address/AppBlock.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: AppBlock.java,v 1.3 1998/12/18 09:53:20 christ Exp $ */

package com.syncbuilder.storage.address;

import java.io.*;
import com.syncbuilder.util.*;

/**
* An AppBlock for the Address-Book application.
*
* @author Kenneth Albanowski, Tilo Christ
*/
public class AppBlock 
	extends com.syncbuilder.storage.CategoryAppBlock 
{
	public AppBlock() 
	{
		super();
	}
		
	public AppBlock(byte[] contents) 
	{
		super(contents);
	}
	
	public void unpack(byte[] data)
	{
		if (data.length < com.syncbuilder.storage.CategoryAppBlock.LENGTH + 4+16*22+2+2)
			throw new java.lang.IllegalArgumentException("data-array too short!");

		super.unpack(data);

		int index = com.syncbuilder.storage.CategoryAppBlock.LENGTH;

		int i;

		int r = PackUtils.getLong(data, index);
		for(i = 0; i < 22; i++)
		{
	    	label_renamed[i] = ((r & (1<<i)) != 0);
		}


		index += 4;
		
		for (i = 0; i < 22; i++)
		{
			label[i] = PackUtils.getZeroTermString(data, index + i*16, null);
		}
			
		index += 16*22;
		
		country = PackUtils.getShort(data, index);
		index += 2;
		sort_by_company = (data[index] != 0);

		for(i = 3; i < 8; i++)
		    phone_label[i - 3] = label[i];
		for(i = 19; i < 22; i++)
		    phone_label[i - 19 + 5] = label[i];
	}
	
	public byte[] pack()
	{
		final int destlen = 4 + 16*22 + 2 + 2;
		byte[] buf = new byte[com.syncbuilder.storage.CategoryAppBlock.LENGTH + destlen];
		super.packCategories(buf);

		int index = com.syncbuilder.storage.CategoryAppBlock.LENGTH;
		int i;
		
		for (i = index; i < index + destlen; i++)
			buf[i] = 0x00;

		int r = 0;

		for(i = 0; i < 22; i++)
		{
		    if (label_renamed[i])
			{
		    	r |= (1<<i);
			}
		}

	    PackUtils.setLong(buf, index, r);
		index += 4;

		for (i = 0; i < 22; i++)
		{
			PackUtils.setZeroTermString(buf, index + i*16, label[i]);
		}
			
		index += 16*22;

		PackUtils.setShort(buf, index, country);
		index += 2;
		buf[index] = (byte)(sort_by_company ? 0xff : 0x00);

		return buf;
	}
	
	public void fill() 
	{
		super.fill();
	
		sort_by_company = false;
		country = 0;
		label = new String[22];
		label_renamed = new boolean[22];
		phone_label = new String[8];
	}
		
	public String describe() 
	{
		StringBuffer c = new StringBuffer("labels=[");
		for(int i=0;i<label.length;i++) 
		{
			if (i>0)
			{
				c.append(", ");
			}
			c.append(label[i]);
		}
		
		c.append("], phone_label=[");
		for(int i=0;i<phone_label.length;i++) {
			if (i>0)
				c.append(", ");
			c.append(phone_label[i]);
		}
		c.append("], label_renamed=[");
		for(int i=0;i<label_renamed.length;i++) {
			if (i>0)
				c.append(", ");
			c.append(label_renamed[i]);
		}
		c.append("], sort_by_company="+sort_by_company+", country="+country); 

		return c.append(", ") + super.describe();
	}

	public boolean sort_by_company;
	public int country;
	public String[] label;
	public String[] phone_label;
	public boolean[] label_renamed;
}
