/* com/syncbuilder/storage/address/Record.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/
/* $Id: Record.java,v 1.3 1998/12/15 09:07:14 christ Exp $ */

package com.syncbuilder.storage.address;

import java.io.*;
import com.syncbuilder.util.*;

/** 
* A representation of an address database record.
*
* @author Kenneth Albanowski, Tilo Christ
*/

public class Record 
	extends com.syncbuilder.storage.Record 
{
	public Record() 
	{
		super();
	}

	public Record(byte[] contents, com.syncbuilder.storage.RecordID id, int index, int attr, int cat) 
	{
		super(contents, id, index, attr, cat);
	}

	public void unpack(byte[] data)
	{
		if (data.length < 9)
			throw new java.lang.IllegalArgumentException("data-array too short!"); 

		// data[0] is gap-fill
	    show_phone     = _hi(data[1]);
  		phone_label[4] = _lo(data[1]);
  		phone_label[3] = _hi(data[2]);
  		phone_label[2] = _lo(data[2]);
  		phone_label[1] = _hi(data[3]);
  		phone_label[0] = _lo(data[3]);
  
	    int contents = PackUtils.getLong(data, 4);
  
		// data[8] is offset of company entry

		int index = 9;
		int len = data.length - 9;
		
		for(int v = 0; v < 19; v++) 
		{
		    if((contents & (1 << v)) != 0) 
			{
		    	if (len < 1)
				{
			        return;
				}
				Index index2 = new Index();
			    entry[v] = PackUtils.getZeroTermString(data, index, index2);
		        index = index2.getValue();
      			len -= entry[v].length() + 1;
    		} 
			else 
			{
      			entry[v] = null;
		    }
  		}	
		
	}
	
	public byte[] pack()
	{
		int destlen = 9;

		int v;
		
		for (v=0; v < 19; v++)
		{
			if (entry[v] != null)
			{
				destlen += entry[v].length() + 1;
			}
		}	
		
		byte[] buf = new byte[destlen];

		int index = 9;
		
		int phoneflag = 0;
	    int contents = 0;
  		int offset = 0;

		for(v = 0; v < 19; v++) 
		{
		    if((entry[v] != null) && (entry[v].length() > 0)) 
			{
	        	if(v == 2)  // 2 == company entry
				{
			        offset = (byte)(index - 8);
				}

			    contents |= (1 << v);
      			index = PackUtils.setZeroTermString(buf, index, entry[v]);
    		}
  		}
  
		phoneflag  = phone_label[0] << 0;
  		phoneflag |= phone_label[1] << 4;
  		phoneflag |= phone_label[2] << 8;
  		phoneflag |= phone_label[3] << 12;
  		phoneflag |= phone_label[4] << 16;
  		phoneflag |= show_phone << 20;

  		PackUtils.setLong(buf, 0, phoneflag);
  		PackUtils.setLong(buf, 4, contents);
  		buf[8] = (byte)offset;

		return buf;
	}

	public void fill() 
	{
		super.fill();
	
		entry = new String[19];
		phone_label = new int[5];
		show_phone = 0;
	}

	/**
	* Describe the Record without any knowledge about the AppBlock.
	*/
	public String describe()
	{
		return describe(null);
	}

	/**
	* Describe the Record with knowledge about the AppBlock (replace numbers with labels).
	*/    
	public String describe(com.syncbuilder.storage.address.AppBlock appblock) 
	{
    	StringBuffer c = new StringBuffer(super.describe((com.syncbuilder.storage.CategoryAppBlock)appblock));

		if (isDeleted())
		{
			c.append("  !!!DELETED!!!");
			return c.toString();
		}

        c.append(", \nphone_label = [");
        for(int i = 0; i < phone_label.length; i++) 
		{
        	if (i > 0)
			{
        		c.append(", ");
			}
        	if (appblock == null)
			{
				c.append(phone_label[i]);
			}
			else
			{
				c.append(appblock.phone_label[phone_label[i]]);
			}
        }
        c.append("], \nentry = [");

        for(int i = 0; i < entry.length; i++) 
		{
        	if (i > 0)
			{
        		c.append(", ");
			}
			
        	c.append((entry[i] == null) ? "(null)" : entry[i]);
        }
        c.append("]");
		
		if (appblock == null)
		{
		    c.append(", \nshow_phone = " + show_phone);
		}
		else
		{
			c.append(", \nshow_phone = " + appblock.phone_label[show_phone]);
		}
		
		return c.toString();
    }


	/**
	* Check two Records for equality. This is mostly for testing purposes.
	*/
	public boolean equals(Object obj)
	{
		if (obj == null)
			return false;

		if (!(obj instanceof com.syncbuilder.storage.address.Record))
			return false;

		if (super.equals(obj) == false)
			return false;

		com.syncbuilder.storage.address.Record other = (com.syncbuilder.storage.address.Record)obj;

		if (show_phone != other.show_phone)
			return false;

		int i;

		for (i = 0; i < 5; i++)
		{
			if (phone_label[i] != other.phone_label[i])
				return false;
		}

		for (i = 0; i < 19; i++)
		{		
			if (entry[i] == null && other.entry[i] != null)
				return false;

			if (entry[i] != null && !entry[i].equals(other.entry[i]))
				return false;
		}
					
		return true;
	}


	/** Which types of phone numbers are there (index into the phone_label String-array of the AppBlock) */
	public int[] phone_label;

	/** Show which phone number as the primary one (index into the phone_label String-array of the AppBlock) */
	public int show_phone;

	/** The entries, which may be <code>null</code> */
	public String[] entry;


	private final static int _hi(byte value)
	{
		return (value >> 4) & 0x0F;
	}

	private final static int _lo(byte value)
	{
		return value & 0x0F;
	}
}
