/* com/syncbuilder/storage/mail/Record.java:
*
* Copyright (C) 1997, 1998, Kenneth Albanowski
* Copyright (C) 1998, Tilo Christ
* Copyright (C) 1998, Robert Crawford
*
* This is free software, licensed under the GNU Library Public License V2.
* See the file COPYING.LIB for details.
*/

package com.syncbuilder.storage.mail;

import java.io.*;
import com.syncbuilder.util.*;

/**
* A representation of a mail database record.
*
* @author Kenneth Albanowski, Tilo Christ, Robert Crawford
*/
public class Record 
	extends com.syncbuilder.storage.Record 
{
	public Record() 
	{
		super();
	}

	public Record(byte[] contents, com.syncbuilder.storage.RecordID id, int index, int attr, int cat) 
	{
		super(contents, id, index, attr, cat);
	}

	public void unpack(byte[] data) {
		int index = 0;


		// unpack date
		int year = (int) ((data[0] & 0xfe) >> 1);
		year += 4;
		int month = (int) ( ( (data[0] & 0x01) << 2) |
					((data[1] & 0xE0) >> 5) );
		int day = (int) (data[1] & 0x1f);
		int hour = (int) data[2];
		int min = (int) data[3];
		date = new java.util.Date (year, month, day, hour, min);
		
		index += 4;

		// read flags;
		byte flags = data[index];
		read = (flags & 0x80) != 0;
		signature = (flags & 0x40) != 0;
		confirmRead = (flags & 0x20) != 0;
		confirmDelivery = (flags & 0x10) != 0;

		priority = (flags & 0x0C) >> 2;
		addressing = (flags & 0x03);

		index = 6;

		// read the subject
		Index idx = new Index();
		subject = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		from = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		to = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		cc = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		bcc = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		replyTo = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		sentTo = PackUtils.getZeroTermString(data, index, idx);
		index = idx.getValue();
		body = PackUtils.getZeroTermString(data, index, idx);
		}

	public byte[] pack() {
		int length = 6;
		length += subject.length() + 1;
		length += from.length() + 1;
		length += to.length() + 1;
		length += cc.length() + 1;
		length += bcc.length() + 1;
		length += replyTo.length() + 1;
		length += sentTo.length() + 1;
		length += body.length() + 1;

		byte[] data = new byte[length];
		int index = 0;

		// pack date
		java.util.Calendar cal = java.util.Calendar.getInstance();
		cal.setTime(date);
		int year = cal.get(java.util.Calendar.YEAR);
		year -= 1904; // Mac-style year
		int month = cal.get(java.util.Calendar.MONTH);
		month++; // correct zero-offset
		int day = cal.get(java.util.Calendar.DATE);
		int hour = cal.get(java.util.Calendar.HOUR_OF_DAY);
		int min = cal.get(java.util.Calendar.MINUTE);

		byte first_byte, second_byte;

		first_byte = (byte) year;
		first_byte = (byte) (first_byte << 1);
		first_byte |= (byte) (month >> 3);
		data[0] = first_byte;
		second_byte = (byte) ((month & 0x03) << 5);
		second_byte |= ( ((byte) day) & 0x1f);
		data[1] = second_byte;
		data[2] = (byte) hour;
		data[3] = (byte) min;

		// pack flags
		byte flags = 0x00;
		flags |= read ? 0x80 : 0x00;
		flags |= signature ? 0x40 : 0x00;
		flags |= confirmRead ? 0x20 : 0x00;
		flags |= confirmDelivery ? 0x10 : 0x00;

		flags |= ((priority & 0x03) << 2);
		flags |= (addressing & 0x03);

		data[4] = flags;

		// pack strings
		index = 6;
		index = PackUtils.setZeroTermString(data, index, subject);
		index = PackUtils.setZeroTermString(data, index, from);
		index = PackUtils.setZeroTermString(data, index, to);
		index = PackUtils.setZeroTermString(data, index, cc);
		index = PackUtils.setZeroTermString(data, index, bcc);
		index = PackUtils.setZeroTermString(data, index, replyTo);
		index = PackUtils.setZeroTermString(data, index, sentTo);
		index = PackUtils.setZeroTermString(data, index, body);
		
		return data;
	}

	public void fill() 
	{
		super.fill();
		read = false;
		signature = false;
		confirmRead = false;
		confirmDelivery = false;
		priority = 0;
		addressing = 0;
		date = new java.util.Date();
		subject = "";
		from = "";
		to = "";
		cc = "";
		bcc = "";
		replyTo = "";
		sentTo = "";
		body = "";
	}

    public String describe() 
	{
        StringBuffer c = new StringBuffer("read=");
      return "read="+read+", signature="+signature+", confirmRead="+confirmRead+
             ", confirmDelivery="+confirmDelivery+", priority="+priority+
             ", addressing="+addressing+", date="+((date == null)?"<Null":date.toString())+
             ", from="+com.syncbuilder.util.Util.prettyPrint(from)+
             ", to="+com.syncbuilder.util.Util.prettyPrint(to)+
             ", cc="+com.syncbuilder.util.Util.prettyPrint(cc)+
             ", bcc="+com.syncbuilder.util.Util.prettyPrint(bcc)+
             ", replyTo="+com.syncbuilder.util.Util.prettyPrint(replyTo)+
             ", sentTo="+com.syncbuilder.util.Util.prettyPrint(sentTo)+
             ", body="+com.syncbuilder.util.Util.prettyPrint(body)+
             ", subject="+com.syncbuilder.util.Util.prettyPrint(subject)+
             ", "+super.describe();
    }

	public boolean read, signature, confirmRead, confirmDelivery;
	public int priority, addressing;

	public java.util.Date date;
	public String subject, from, to, cc, bcc, replyTo, sentTo, body;
}
